/*
 * character.h
 *
 *  Created on: 23-08-2013
 *      Author: crm
 */

#pragma once

#include <list>

#include "asmmath.h"
#include "lua2.h"

#include "graphics/sprite.h"
#include "graphics/spriteptr.h"

#include "combomanager.h"
#include "idea.h"
#include "objecttype.h"
#include "base/controller.h"
#include "base/event.h"
#include "base/serializable.h"
#include "base/eventprovider.h"

class GameMode;

class Character: public Serializable, public ObjectType, public EventProvider
	{
	public:
		enum Direction
			{
			LEFT=Graphics::LEFT,
			RIGHT=Graphics::RIGHT
			};

		enum State
			{
			NONE=			0x0,
			CONTROLLABLE=	0x1<<0,
			NOCHANGE=		0x1<<1,
			MOVING=			0x1<<4,
			DEFENDING=		0x1<<5,
			CASTING=		0x1<<6,	// Ladowanie pomyslu
			ATTACKING=		0x1<<7,	// Trzyma pomysl w rece
			COMBO=			0x1<<8,
			DEAD=			0x1<<9,
			DISABLED=		0x1<<10,
			};

		struct Stats
			{
			Stats(): maxHP(1.0f), maxMP(1.0f), speed(100.0f), castspeed(1.0f), strength(1000.0f) {}

			float maxHP;			// Maksymalna liczba zyc
			float maxMP;			// Maksymalna liczba energii
			float speed;			// Maksymalna szybkosc
			float castspeed;		// Szybkosc castowania (mana na sekunde)
			float strength;			// Sila rzucania pomyslami
			};

	protected:
		AVector pos;
		AVector spd;
		Direction dir;

		Controller *ctrl;
		GameMode *gamemode;
		Stats stats;
		Lua lua;
		ComboManager combo;

		std::vector<Idea *> ideas;

		unsigned state;

		int id;

		float hp;			// Zycia
		float mp;			// Zdolnosc mowienia

		float castusage;	// Ilosc powietrza uzyta do aktualnej idei
		int castid;			// Aktualnie castowany pomysl
		int casted;			// Zaladowany pomysl (-1 -> brak)

		Idea* idea;			// Trzymany pomysl

		float block;		// Czas blokady postaci
		float castblock;	// Czas blokady mowienia
		int ref;			// Referencja na funkcje lua do wykonania. Domyslnie LUA_UPDATE
		float luaacc;

		Event event;
		//std::vector<EventListener *> listeners;

		Graphics::SpritePtr spr;
		Graphics::SpritePtr shadow;
		Graphics::SpritePtr speech;

		std::vector<std::string> sounds;

		std::string name;	// Nazwa postaci
		std::string path;	// Sciezka do pliku postaci

		struct CharacterSerialized // 48b
			{
			AVector pos;
			AVector spd;
			Direction dir;
			unsigned state;
			float hp;
			float mp;
			};

		inline void addState(int flag) {state|=flag;}
		inline void delState(int flag) {state&=~flag;}

		void updateCtrls(float dt);
		void updateMove(float dt);
		void updateAnim(float dt);
		void updateMp(float dt);

		void updateClient(float dt);

		float getCostSum() const {float s=0.0f; for(unsigned i=0u; i<ideas.size(); ++i) s+=ideas[i]->getCost(); return s;}

	public:
		Character(): Serializable(), ObjectType(), pos(), spd(), dir(RIGHT), ctrl(NULL), gamemode(NULL),
			combo(), state(NONE), id(-1), hp(1.0f), mp(0.0f), castusage(0.0f), castid(0), casted(-1), idea(NULL), block(0.0f), castblock(0.0f),
			ref(0), luaacc(0.0f), event(0, 0, 0, 0), spr(), shadow(), speech() {}
		virtual ~Character() {clear();}

		virtual bool init(const std::string& script);
		virtual void update(float dt);
		virtual void print();
		virtual void clear();

		virtual char* serialize(int &size);
		virtual void deserialize(char *buff);

		//void addListener(EventListener *lst) {listeners.push_back(lst);}
		void pushEvent(unsigned e, float dataf=0.0f, int datai=0, void *datap=0);
		void sendEvent(const Event& e);

		void addCombo(const ComboManager::Combo& c) {combo.addCombo(c);}
		//void startCombo() {ref=combo.getRef(); addState(COMBO); pushEvent(Event::CHARACTER_COMBO, ref);}
		bool startCombo(int sref);// {ref=sref; addState(COMBO); pushEvent(Event::CHARACTER_COMBO, ref);}
		void  stopCombo() {delState(COMBO); ref=0;}

		void addIdea(Idea *i) {ideas.push_back(i);}	// Dodaje pomysl DO LISTY POMYSLOW!
		void giveIdeaSelf(int id);
		void giveIdeaOther(int id);

		void disableControl() {addState(DISABLED);}
		void  enableControl() {delState(DISABLED);}
		void revive();

		bool canCast() const;
		bool canDefend() const;

		virtual void attack(bool noevent=false);
		virtual void defend();
		virtual void hit(float dmg);

		void moveLeft();
		void moveRight();
		void moveUp();
		void moveDown();
		void jump(float force);

		inline void setPos(AVector s) {pos=s; pushEvent(Event::CHARACTER_MISC);}
		inline void setSpd(AVector s) {spd=s; pushEvent(Event::CHARACTER_MISC);}
		inline void setDir(Direction s) {dir=s; pushEvent(Event::CHARACTER_MISC);}
		inline void setController(Controller *s) {ctrl=s;}
		inline void setGameMode(GameMode *s) {gamemode=s;}
		inline void setState(unsigned s) {state=s;}
		inline void setId(int s) {id=s;}
		inline void setHp(float s) {hp=(s>stats.maxHP)?stats.maxHP:s; pushEvent(Event::CHARACTER_MISC);}
		inline void setMp(float s) {mp=(s>stats.maxMP)?stats.maxMP:s; pushEvent(Event::CHARACTER_MISC);}
		//inline void setUse(float s) {use=(s>mp)?mp:s;}
		inline void setBlock(float seconds) {block=seconds;}
		inline void setCastBlock(float seconds) {castblock=seconds;}
		inline void setLuaRef(int s) {ref=s; if(ref>0) addState(COMBO); pushEvent(Event::CHARACTER_MISC);}	// Zostawic czy wywalic dodawanie state combo? Teoretycznie dla kazdego >0 bedzie to chyba combo-funkcja
		inline void setSprite(const std::string& name) {spr.setSprite(name);}
		inline void setSpriteShadow(const std::string& name) {shadow.setSprite(name);}
		inline void setSpriteSpeech(const std::string& name) {speech.setSprite(name);}
		//inline void setId(int s) {id=s;}
		inline void setName(const std::string& s) {name=s;}
		inline void setStatsMaxHp(float s) {stats.maxHP=s;}
		inline void setStatsMaxMp(float s) {stats.maxMP=s;}
		inline void setStatsMovementSpeed(float s) {stats.speed=s;}
		inline void setStatsCastSpeed(float s) {stats.castspeed=s;}
		inline void setStatsStrength(float s) {stats.strength	=s;}
		//inline void setStatsVoiceVolume(float s) {stats.voiceVolume=s;}
		//inline void setStatsSpeakSpeed(float s) {stats.speakSpeed=s;}
		//inline void setStatsLungCapacity(float s) {stats.lungCapacity=s;}
		//inline void setStatsAirUsage(float s) {stats.airUsage=s;}
		inline void setAnim(const std::string& anim) {spr.setAnim(anim);}

		inline AVector getPos() const {return pos;}
		inline AVector getActPos() const {int ax, ay; spr.getActionPoint(ax, ay); return AVectorAdd(pos, AVectorSet(ax, ay, 0.0f, 0.0f));}
		inline AVector getSpd() const {return spd;}
		inline Direction getDir() const {return dir;}
		inline GameMode* getGameMode() const {return gamemode;}
		inline const Stats& getStats() const {return stats;}
		inline unsigned getState() const {return state;}
		inline bool getState(int flag) const {return (state&flag);}
		inline float getHp() const {return hp;}
		inline float getMp() const {return mp;}
		inline float getUse() const {return castusage;}
		inline float getBlock() const {return block;}
		inline float getCastBlock() const {return castblock;}
		inline int getCombo() const {return ref;}
		inline int getId() const {return id;}
		inline Lua& getLua() {return lua;}

		inline unsigned addSound(const std::string& name) {sounds.push_back(name); return sounds.size()-1u;}
		inline const std::string& getSound(unsigned i) {static std::string n; return (i<sounds.size())?sounds[i]:n;}

		inline const std::string& getName() const {return name;}
		inline const std::string& getPath() const {return path;}
	};

