/*
 * Camera.cpp
 *
 *  Created on: 07-03-2013
 *      Author: crm
 */

#include "camera.h"

#include <cmath>
#include <GL/glee.h>
#include <GL/gl.h>

#include "engine.h"
#include "render.h"
#include "log.h"

using namespace Render;

void Camera::update()
	{
	const AMatrix modelview=AMatrixTranspose(getModelview());	// Transpozycja bo opengl chce macierz zapisana kolumnami
	const AMatrix projection=AMatrixTranspose(getProjection());	// ... a u mnie jest wierszami

	if(depthbuffer)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);

	glViewport(cx, cy, cw, ch);

	switch(type)
		{
		case ORTHOGRAPHIC:
		case PERSPECTIVE:
			//glOrtho(-cw/2, cw/2, -ch/2, ch/2, cutnear, cutfar);
			//glFrustum(-fw, fw, -fh, fh, cutnear, cutfar);

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glMultMatrixf((float*)&projection);

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			glMultMatrixf((float*)&modelview);
			glTranslatef(-pos.x, -pos.y, -pos.z);
		break;

		default:
		case GUI:
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glMultMatrixf((float*)&projection);

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
		break;
		}
	}

AMatrix Camera::getModelview() const
	{
	return AMatrix(right, up, -forward, AVector(0, 0, 0, 1));// - AMatrix(AVector(0, 0, 0, pos.x), AVector(0, 0, 0, pos.y), AVector(0, 0, 0, pos.z), AVector(0, 0, 0, 0));
	}

AMatrix Camera::getProjection() const
	{
	const float N=cutnear;
	const float F=cutfar;

	switch(type)
		{
		case ORTHOGRAPHIC:
			{
			const float R= cw/2;
			const float L=-cw/2;
			const float T= ch/2;
			const float B=-ch/2;
			const float TX=-(R+L)/(R-L);
			const float TY=-(T+B)/(T-B);
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector( 2/(R-L), 0, 0, TX),
					AVector(0,  2/(T-B), 0, TY),
					AVector(0, 0, -2/(F-N), TZ),
					AVector(0, 0, 0, 1)
					);
			}
		break;

		case PERSPECTIVE:
			{
			const float ASPECT=cw/ch;
			const float FH=tan(fov/360.0*M_PI);
			const float FW=FH*ASPECT;
			const float C=-(F+N)/(F-N);
			const float D=-(2*F*N)/(F-N);

			return AMatrix(
					AVector(1/FW, 0, 0, 0),
					AVector(0, 1/FH, 0, 0),
					AVector(0, 0,  C, D),
					AVector(0, 0, -1, 0)
					);
			}
		break;

		default:
		case GUI:
			{
			const float R=cw;
			const float L=0;
			const float T=ch;
			const float B=0;
			const float TX=-(R+L)/(R-L);
			const float TY=-(T+B)/(T-B);
			const float TZ=-(F+N)/(F-N);

			return AMatrix(
					AVector( 2/(R-L), 0, 0, TX),
					AVector(0,  2/(T-B), 0, TY),
					AVector(0, 0, -2/(F-N), TZ),
					AVector(0, 0, 0, 1)
					);
			}
		break;
		}

	return AMatrix(); // i tak tu nie dojdzie...
	}

Orientation Camera::getBillboard(const AVector& objpos) const
	{
	switch(type)
		{
		case ORTHOGRAPHIC:
			{
			const AVector f=AVectorNormalize(AVector(forward.x, forward.y, 0.0f));
			//const float ce=AVectorDot(up, AVector(0, 0, 1)).w;
			const float ce=AVectorDot(forward, f).w;

			//printf("ce %f", ce);

			if(ce<0.1f && ce>-0.1f)
				return Orientation(AVector(), AVector(0, 0, 1), -f, AVector(1, 0.0f, 1));
			return Orientation(AVector(), AVector(0, 0, 1), -f, AVector(1, 1.0f/ce, 1));
			}
		break;

		case PERSPECTIVE:
			{
			const AVector dist=objpos-pos;
			const AVector f=AVectorNormalize(AVector(dist.x, dist.y, 0.0f));

			return Orientation(AVector(), AVector(0, 0, 1), -f, AVector(1, 1, 1));
			}
		break;

		default:
		case GUI:
			return Orientation::GUI;
		break;
		}
	}

AVector Camera::getPointOnScreen(const AVector& objpos) const
	{
	AVector ret;

	ret=objpos-pos;
	ret=getModelview()*ret;
	ret=getProjection()*ret;
	if(type==PERSPECTIVE)
		ret=ret/AVector(ret.z, ret.z, 1, 1);
	ret.x=(ret.x+1)*cw/2;
	ret.y=(ret.y+1)*ch/2;
	ret.z=0;
	ret.w=0;

	return ret;
	}

void Camera::getRay(int mx, int my, AVector& raypos, AVector& raydir) const
	{
	float x;
	float y;

	if(type==PERSPECTIVE)
		{
		const float aspect=(float)cw/ch;
		float fw, fh;
		fh=tan(fov/360.0*M_PI);
		fw=fh*aspect;

		x=( mx/cw-0.5f)*2.0f;
		y=(-my/ch+0.5f)*2.0f;

		x=fw*x;
		y=fh*y;

		raypos=pos + forward*cutnear + right*x + up*y;
		raydir=AVectorNormalize(raypos-pos);
		}
	else
		{
		x=( mx/(float)cw-0.5f)*cw*zoom;
		y=(-my/(float)ch+0.5f)*ch*zoom;

		raypos=pos - forward*cutnear + right*x + up*y;
		raydir=forward;
		}
	}
