#include "resfile.h"

#include <cstring>
#include <zlib.h>

#include "log.h"
#include "parser.h"

const char FILESTAMP[]="CRES";							// Typ pliku
const char VERSION=2;									// Wersja druga mojego cudownego resfile

bool Resfile::open(const std::string& res)
	{
	/*if(res.compare(curr)==0)
		return true;*/

	clear();

	in.open(res.c_str(), std::ios::in|std::ios::binary);

	if(!in)
		return false;

	//curr=res;

	/**** Sprawdzenie naglowka i wersji ****/
		{
		const unsigned IDENT_LENGTH=strlen(FILESTAMP);

		char ident[IDENT_LENGTH+1];
		char version;

		in.seekg(0, std::ios::end);
		rsize=in.tellg();
		in.seekg(0, std::ios::beg);

		if(rsize<IDENT_LENGTH+1u)	// + bajt na wersje
			return false;

		in.read(ident, IDENT_LENGTH);
		in.read(&version, 1);

		if(in.fail())
			return false;

		ident[IDENT_LENGTH]='\0';

		if(strcmp(ident, FILESTAMP)!=0)
			return false;
		if(version!=VERSION)
			return false;
		}

	/**** Wczytanie naglowka ****/
	unsigned hsize, count;
	in.read(reinterpret_cast<char*>(&hsize), sizeof(unsigned));
	in.read(reinterpret_cast<char*>(&count), sizeof(unsigned));
	header.reserve(count);

	Item item;
	unsigned nlen;
	std::vector<char> name;

	for(unsigned int i=0; i<count; i++)
		{
		in.read((char*)&item.start, sizeof(unsigned));
		in.read((char*)&item.size, sizeof(unsigned));
		in.read(&item.flags, sizeof(char));

		if(item.flags&COMPRESSED)
			in.read((char*)&item.ucsize, sizeof(unsigned));
		else
			item.ucsize=item.size;

		in.read(reinterpret_cast<char*>(&nlen), sizeof(unsigned));
		name.resize(nlen+1);
		in.read(name.data(), nlen);
		name[nlen]='\0';

		int fname;
		for(fname=nlen-1; fname>=0 && name[fname]!='/'; --fname)

		item.path=name.data();
		item.name=name.data()+fname+1;
		item.start+=hsize;

		//Log::debug("Resource.open: \"%s\" (%s); start %u, size %u, ucsize %u, flags 0x%02X", item.path.c_str(), item.name.c_str(), item.start, item.size, item.ucsize, item.flags);

		if(in.fail())
			return false;

		header.push_back(new Item(item));
		treeAddItem(item.path, header.back());
		}

	size=0u;

	return true;
	}

void Resfile::treeAddItem(const std::string& path, Item* item)
	{
	Parser pars;
	TreeItem *leaf;
	TreeItem nleaf;

	pars.setDelimiter("/");
	pars.setText(path);

	leaf=&root;

	for(int i=0; i<pars.count()-1; ++i)
		{
		TreeItem *tmp=treeFindDirectory(pars[i], leaf);
		if(tmp)
			leaf=tmp;
		else
			{
			nleaf.name=pars[i];
			nleaf.type=TreeItem::DIRECTORY;
			leaf->items.push_back(nleaf);
			leaf=&leaf->items.back();
			}
		}

	nleaf.name=pars[pars.count()-1];
	nleaf.type=TreeItem::FILE;
	nleaf.item=item;
	leaf->items.push_back(nleaf);
	}

Resfile::TreeItem* Resfile::treeFind(const std::string& name, TreeItem* leaf, TreeItem::Type type)
	{
	for(unsigned i=0; i<leaf->items.size(); ++i)
		{
		if(leaf->items[i].type!=type)
			continue;
		if(leaf->items[i].name.compare(name)==0)
			return &leaf->items[i];
		}

	return NULL;
	}

void Resfile::clear()
	{
	if(in.is_open())
		in.close();

	//curr=std::string();
	size=0u;
	rsize=0u;

	for(std::vector<Item *>::iterator it=header.begin(); it!=header.end(); ++it)
		delete *it;
	header.clear();

	root.type=TreeItem::DIRECTORY;
	root.name=std::string();
	root.items.clear();
	}

char* Resfile::load(Item *entry)
	{
	char *data;
	char *cdata;

	try
		{
		data=new char[entry->size+1];
		}
	catch(const std::bad_alloc &)
		{
		return NULL;
		}

	in.seekg(entry->start);
	in.read(data, entry->size);
	size=entry->size;

	if(entry->flags&COMPRESSED)
		{
		int ret;

		cdata=data;
		size=entry->ucsize;

		try
			{
			data=new char[entry->ucsize+1];
			}
		catch(const std::bad_alloc& e)
			{
			delete [] cdata;
			return NULL;
			}

		ret=uncompress((Bytef*)data, (uLongf*)&entry->ucsize, (Bytef*)cdata, entry->size);
		delete [] cdata;

		switch(ret)
			{
			case Z_MEM_ERROR:
				Log::fail("Resfile.load: Za malo pamieci do rozpakowania pliku \"%s\"", entry->name.c_str());
			break;

			case Z_BUF_ERROR:
				Log::fail("Resfile.load: Bufor wyjsciowy jest za maly \"%s\"", entry->name.c_str());
			break;

			case Z_DATA_ERROR:
				Log::fail("Resfile.load: Problem z rozpakowaniem danych wejsciowych \"%s\"", entry->name.c_str());
			break;

			default:
			case Z_OK:
				//Log::ok("Resfile.load: Rozpakowano plik \"%s\" (bylo %u, jest %u)", entry->name.c_str(), entry->size, entry->ucsize);
			break;
			}

		if(ret!=Z_OK)
			{
			delete [] data;
			return NULL;
			}
		}

	data[size]='\0';

	//Log::debug("Resfile.load: Wczytano plik \"%s\", wielkosc: %u; skompresowany: %d (wielkosc po dekompresji: %u)", entry->name.c_str(), entry->size, entry->flags&COMPRESSED, entry->ucsize);

	return data;
	}

Resfile::Item* Resfile::findItem(const std::string& path)
	{
	/*for(unsigned i=0; i<header.size(); ++i)
		if(header[i]->path.compare(path)==0)
			return header[i];
	return NULL;*/

	Parser pars;
	pars.setDelimiter("/");
	pars.setText(path);

	TreeItem *leaf=&root;

	for(int i=0; i<pars.count()-1; ++i)
		{
		leaf=treeFindDirectory(pars[i], leaf);

		if(!leaf)
			return NULL;
		}

	leaf=treeFindFile(pars[pars.count()-1], leaf);

	if(leaf)
		return leaf->item;
	return NULL;
	}

char* Resfile::load(const std::string& res, const std::string& file)
	{
	//Item *item;
	if(!open(res))
		return NULL;

	/*item=findItem(file);

	if(item)
		return load(item);
	return NULL;*/
	return load(file);
	}

char* Resfile::load(const std::string& path)
	{
	/*Parser pars;

	pars.setDelimiter(":");
	pars.setText(path);

	if(pars.count()<2)
		return load(curr, path);
	return load(pars[0], pars[1]);*/

	Item *item;

	item=findItem(path);

	if(item)
		return load(item);
	return NULL;
	}

char* Resfile::data(const std::string& res)
	{
	rsize=0u;
	if(!open(res))
		return NULL;

	char *data;

	try
		{
		data=new char[rsize];
		}
	catch(const std::bad_alloc &)
		{
		return NULL;
		}

	std::fstream in;

	in.open(res.c_str(), in.in|in.binary);
	in.read(data, rsize);
	in.close();

	size=rsize;

	return data;
	}


bool Resfile::find(const std::string& path)
	{
	Parser pars;

	pars.setDelimiter(":");
	pars.setText(path);

	if(pars.count()>1)
		{
		open(pars[0]);
		return findItem(pars[1]);
		}

	return findItem(path);
	}


unsigned Resfile::getSize(const std::string& path)
	{
	Parser pars;
	Item *item;

	pars.setDelimiter(":");
	pars.setText(path);

	if(pars.count()>1)
		{
		open(pars[0]);
		item=findItem(pars[1]);
		}
	else
		item=findItem(path);

	if(item)
		return item->ucsize;
	return 0u;
	}


const std::string& Resfile::getFileName(unsigned int num)
	{
	return header[num]->name;
	}

